// Java script to visualize the standard position error of single point determinations 
// based on Bachelor’s Thesis of Lukas Raffl (2016)
//
// uses Numeric Javascript library (1.2.6), Sébastien Loisel, http://www.numericjs.com/
//
// Technical University of Munich
// Chair of Geodesy
// L. Raffl
// 16.08.2016


function fitToWindow(){	
	// Function to fit the canvas elements to the window size
	
	// Resize the drawing area
	canvas = document.getElementById("screen");
	canvas.style.width = '100%';
	canvas.style.height = '100%';
	// the ratio width:heigth = 1:0.7 
	var oldWidth = canvas.width;
	canvas.width  = canvas.offsetWidth;
	canvas.height = canvas.width * 0.7;
	
	// calculate the coordinates of the center and the new scaling factor
	var yM1 = oldWidth/2;
	var xM1 = yM1 * 0.7;	
	var yM2 = canvas.width/2;
	var xM2 = yM2 * 0.7;
	var m = yM2/yM1;
	
	// Resize the canvas for the scalebars
	canvas = document.getElementById("scale");
	canvas.style.width = '100%';
	canvas.style.height = '100%';
	canvas.width  = canvas.offsetWidth;
	// the canvas heigth is at least 60px
	canvas.height = Math.max(canvas.width * 0.07, 60);
	
	// Resize the buttonboxes
	var leftBox = document.getElementById("creatingTools");
	var rightBox = document.getElementById("deletingTools");
	
	var wLeft = leftBox.offsetWidth;
	// if the window size is too small, align the boxes vertically
	if(canvas.width < 480){
		leftBox.style.width = '98%';
		wLeft = leftBox.offsetWidth;
		rightBox.style.width = '96%'
		rightBox.style.paddingLeft = '2%';
		rightBox.style.paddingRight = '2%';
		rightBox.style.float = 'left';
	// if the window size is big enough, align the boxes horizontally
	} else {
		leftBox.style.width = '70%';
		wLeft = leftBox.offsetWidth;
		rightBox.style.width = '112px';
		// calculate the padding that the box size is 26%
		rightBox.style.paddingLeft = (26/72*wLeft - 112)*0.48 + 'px';
		rightBox.style.paddingRight = (26/72*wLeft - 112)*0.52 + 'px';
		rightBox.style.float = 'right';
	}
	
	// Resize the sliders
	// in order to get the new width of the sliders, the width of the button and the text box has to be subtracted
	var wSlider = 0.92*wLeft - 130 - 80;
	document.getElementById("slider1").style.width = wSlider + 'px';
	document.getElementById("slider2").style.width = wSlider + 'px';
	document.getElementById("slider3").style.width = wSlider + 'px';
	
	// transform the points, so that they fit to the new canvas size
	if(points == undefined){return};	
	var Q1 = numeric.rep([points.length], [yM1, xM1]);
	var Q2 = numeric.rep([points.length], [yM2, xM2]);		
	points = numeric.add(Q2, numeric.mul(m, numeric.sub(points, Q1)));

	// All calculations are based on a local pixel coordinate system. So when the map scale is changed, also the angle measurement accuracy has to be changed, because the points are transformed and the distances between the points in the pixel coordinate system are changing, but the simulated distances and accuracies should remain the same.
	sA = sA / m;		
	mapScale = mapScale * m;
	
	// Update graphics
	drawScalebar(mapScale);
	drawGeometry();

}
	  
function mouseClick(e){
	// Function that is called on mouse click
			
	// position of mouse click
	var y = e.pageX - this.offsetLeft;
	var x = e.pageY - this.offsetTop;
	
	// get type of mouse click
	b = e.button;
	
	// on left mouse click
	if(b == 0){
	
		// add control point
		if(command == 1){
			addPoint(y,x);
			buttonClick(0); // cancel activated function
			
		// add measurement or delete point
		} else if(command > 1){
		
			// if no point clicked
			if(pID < 0){
				panMap(y,x);				
				
			// if point clicked
			} else {
				clickOnPoint(command);
			}
			
		// no function is activated
		} else {
		
			// if no point clicked
			if(pID < 0){
				panMap(y,x);
				
			// if point clicked
			} else {
				movePoint(y,x,pID);
			}
			
		}
		
	// ohter type of mouse click
	} else {
		// cancel command
		buttonClick(0);
	}
	
}

function mouseMove(e){
	// Function that is called on mouse move
		
	// current position of the mouse
	var y = e.pageX - this.offsetLeft;
	var x = e.pageY - this.offsetTop;

	// check if mouse is over a point and get its id (-1 if mouse is not over a point)
	var id = selectPoint([y,x], points, 10);
	
	// If the mouse is over a point, the parameters of its error ellipse are shown.
	var tooltip;
	
	// check if the mouse is over a new object
	if(id != pID){
		
		// if the mouse has left a point, no tooltip is shown
		if(id < 0){
			map.title = "";
			drawGeometry(y,x);
			
		// if the mouse has reached a point, a tooltip with the ellipse parameters is created
		} else {
			
			// For an incomplete measurement the current point is nominated as the next point of the measurement.
			drawGeometry(points[id][0], points[id][1]);
			
			// The current point is highlighted in red
			if(id == 0){
				tooltip = "Neupunkt";
				if(command != 4){ // don't highlight the new point when the deleting function is activated, because the new point can't be deleted.
					plot([points[id][0]], [points[id][1]], 'ro');
				}
			} else {
				tooltip = "A = 0.0 mm \nB = 0.0 mm \nphi = 0.0000 gon";
				plot([points[id][0]], [points[id][1]], 'r^');
			}
			
			// calculate the parameters of the error ellipse and create a tooltip
			param = calcEllipse(id);		
			if(param != -1){
				var strA = (Math.round(param[0] * 10000)/10).toString();
				if(param[0] > 9.999){
					strA = ">9999";
				}
				var strB = (Math.round(param[1] * 10000)/10).toString();
				if(param[1] > 9.999){
					strB = ">9999";
				}
				var strPhi = (Math.round(param[2] * 200/Math.PI * 10000)/10000).toString();
				
				tooltip = "A = " + strA + " mm \nB = " + strB + " mm \nphi = " + strPhi + " gon";			
			}
			
			map.title = tooltip;
			
		}
		
		// update the current object id
		pID = id;
		
	// if the mouse is not over a point and there exists a incomplete measurement, show mouse position as next point of the measurement
	} else if(id < 0 && distances[distances.length-1][0] + angles[angles.length-1][0] > -2){
		drawGeometry(y,x);
	}
	
}

function buttonClick(c_id){
	// Function to activate a certain command
	//	Input:	- c_id: id of the command to be activated
	
	// deactivate old command
	switch(command){
			case 1:
				document.getElementById("btn_addPoint").className = "btn";
				break;
			case 2:
				document.getElementById("btn_addDistance").className = "btn";
				break;
			case 3:
				document.getElementById("btn_addAngle").className = "btn";
				break;
			case 4:
				document.getElementById("btn_delete").className = "btn";
				break;
		}
		
	// check if new command is equal to the already activated command or the activated command was canceled
	if(c_id == command || c_id == 0){
		command = 0;
		map.style.cursor = "default";
		// delete any incomplete measurements
		distances[distances.length-1] = [-1, -1];
		angles[angles.length-1] = [-1, -1, -1];
		drawGeometry();
		
	// if a new command was selected, activate it
	} else {
		command = c_id;
		switch(c_id){
			case 1: //add a control point
				document.getElementById("btn_addPoint").className = "btn_set";
				map.style.cursor = "crosshair";
				break;
			case 2: //add a distance measurement
				document.getElementById("btn_addDistance").className = "btn_set";
				map.style.cursor = "crosshair";
				break;
			case 3: //add an angle measurement
				document.getElementById("btn_addAngle").className = "btn_set";
				map.style.cursor = "crosshair";
				break;
			case 4: //delete a control point
				document.getElementById("btn_delete").className = "btn_set";
				map.style.cursor = "crosshair";
				break;
			case 5: //delete all
				command = 0;
				map.style.cursor = "default";
				points = [[map.width/2, map.height/2]];
				distances = [[-1, -1]];
				angles = [[-1, -1, -1]];
				Qxx = [];
				drawGeometry();
				break;
		}
	}
}

function key(e){
	// Function that allows keyboard shortcuts
	
	switch(e.which){
		case 13: //cancel			
			buttonClick(0);
			break;
		case 112: //add a control point
			buttonClick(1);
			break;
		case 100: //add a distance measurement
			buttonClick(2);
			break;
		case 97: //add an angle measurement
			buttonClick(3);
			break;
		case 99: //delete a control point
			buttonClick(4);
			break;
		case 110: //delete all
			buttonClick(5);
			break;
	}
}

function errorCalculation(){
	//Function that computes the error calculation
	
	// Number of points and measurements
	var nP = points.length;
	var nC = nP - 1; // number of control points
	// The last row of the measurement vectors are incomplete measurements, so the real number of measurements is one less.
	var nS = distances.length - 1;
	var nT = angles.length - 1;
	
	// Number of measurements to the new point
	var nN = 0;
	
	// Initializing the necessary matrices
	var A, Kbb;
	A = numeric.rep([2*nC+nS+nT, 2*nP],0);
	Kbb = numeric.rep([2*nC+nS+nT, 2*nC+nS+nT],0);
	
	/**********/	
	// Control points
	
	for(var c=0; c<nC; c++){
		
		// add to Design Matrix
		A[2*c][2*c+2] = 1;
		A[2*c+1][2*c+3] = 1;

		// add to covariance matrix
		Kbb[2*c][2*c] = sP * sP;
		Kbb[2*c+1][2*c+1] = sP * sP;
		
	}
	
	/**********/
	// Distance measurements
	
	var iP1, y1, x1, iP2, y2, x2, dx, dy, s12, iS, sigma;
	
	for(var s=0; s<nS; s++){
		
		// Starting point
		iP1 = distances[s][0];
		y1 = points[iP1][0];
		x1 = points[iP1][1];
		
		// Endpoint
		iP2 = distances[s][1];
		y2 = points[iP2][0];
		x2 = points[iP2][1];
		
		// check if the new point ia involved
		if(iP1 == 0 || iP2 == 0){nN++};
		
		// Distance
		dy = y2 - y1;
		dx = x2 - x1;
		s12 = Math.sqrt(dy*dy + dx*dx);
		
		// add to Design Matrix
		iS = 2*nC + s;
		A[iS][2*iP1] = -dy/s12;
		A[iS][2*iP1+1] = -dx/s12;
		A[iS][2*iP2] = dy/s12;
		A[iS][2*iP2+1] = dx/s12;
		
		// add to covariance matrix
		Kbb[iS][iS] = sD*sD;
		
	}
	
	/**********/
	// Angle measurements
	
	var iP1, y1, x1, iP2, y2, x2, iP3, y3, x3, dx12, dy12, s12, dx13, dy13, s13;
	var iT, sigma;
	
	for(var t=0; t<nT; t++){
		
		// stand
		iP1 = angles[t][0];
		y1 = points[iP1][0];
		x1 = points[iP1][1];
		
		// target 1
		iP2 = angles[t][1];
		y2 = points[iP2][0];
		x2 = points[iP2][1];
		
		// target 2
		iP3 = angles[t][2];
		y3 = points[iP3][0];
		x3 = points[iP3][1];
		
		// check if the new point ia involved
		if(iP1 == 0 || iP2 == 0 || iP3 == 0){nN++};
		
		// distances
		dy12 = y2 - y1;
		dx12 = x2 - x1;
		s12 = Math.sqrt(dy12*dy12 + dx12*dx12);
		dy13 = y3 - y1;
		dx13 = x3 - x1;
		s13 = Math.sqrt(dy13*dy13 + dx13*dx13);
		
		// add to Design Matrix
		iT = 2*nC + nS + t;
		A[iT][2*iP1] = - dx13/(s13*s13) + dx12/(s12*s12);
		A[iT][2*iP1+1] = dy13/(s13*s13) - dy12/(s12*s12);
		A[iT][2*iP2] = -dx12/(s12*s12);
		A[iT][2*iP2+1] = dy12/(s12*s12);
		A[iT][2*iP3] = dx13/(s13*s13);
		A[iT][2*iP3+1] = -dy13/(s13*s13);
		
		// add to covariance matrix
		Kbb[iT][iT] = sA * sA;
		
	}
	
	/**********/
	// Calculation
	
	if(nP > 1){
		
		var Pbb, N;
		var A_t = numeric.transpose(A);
		
		// If there are less than 2 measurements to the new point, its accuracy can't be calculated.
		// In this case, only the accuracy of the control points is calculated.
		if(nN < 2){
			for(var a=0; a<A.length; a++){
				if(A[a][0] != 0){
					A[a] = numeric.rep([2*nP],0);
				}
			}
			var A_t = numeric.transpose(A);
			A_t.shift();
			A_t.shift();
			A = numeric.transpose(A_t);
		}
		
		// accuracy calculation
		Pbb = numeric.inv(Kbb);
		N = numeric.dot(A_t, numeric.dot(Pbb,A));
		Qxx = numeric.inv(N);
		
		// If the accuracy of the new point could not be calculated, its accuracy is set to 0
		if(nN < 2){
			Qxx.unshift(numeric.rep([Qxx[0].length], 0));
			Qxx.unshift(numeric.rep([Qxx[0].length], 0));
			var Qxx_t = numeric.transpose(Qxx);
			Qxx_t.unshift(numeric.rep([Qxx_t[0].length], 0));
			Qxx_t.unshift(numeric.rep([Qxx_t[0].length], 0));
			Qxx = numeric.transpose(Qxx_t);
		}
			
	} else {

		Qxx = [[0, 0],[0, 0]];
		
	}
	
}

function addPoint(y,x){
	// Function to add a control point to the point list
	//	Input:	- (y,x) coordinates of the point
	
	points.push([y, x]);
	drawGeometry();
	
}

function drawGeometry(y,x){
	// Function to draw the measurement configuration. All points as well as all distance and angle measurements of the given measurement configuration are drawn. In addition, the error calculation is performed and the resulting error ellipses are drawn.
	//  Input:	- (y,x) mouse coordinates. These can be used to visualize that just a new measurement is created and not all necessary points have been selected yet. If all measurements are complete, the input of this function is empty.
	
	// check if there is anything to draw
	if(points.length == 0){
		return;
	}
	
	// no input
	if(y === undefined){y = 0};
	if(x === undefined){x = 0};
	
	// clear the canvas
	mapCtx.clearRect(0,0,map.width,map.height);
	
	// to get all y-coordinates in one array and all x-coordinates in another array, transpose the point list
	var P = numeric.transpose(points);
	
	// first point is allways the new point
	plot([P[0].shift()], [P[1].shift()], "o");
	
	// all the other points are control points
	plot(P[0], P[1], "^");
	
	// visualize the angle measurements
	var iM
	for(var t = 0; t < angles.length; t++){
		
		// current measurement
		iM = angles[t];
		
		// Check if the current measurement is complete. If the measurement is incomplete (less than 3 points), the current mouse position is temporary used as next input point for the measurement.
		
		// An empty measurement is not drawn
		if(iM[0] > -1){
			
			// Complete angle measurement inputs will be shown blue, incomplete inputs will be red.
			var color = "b";
			
			// If the measurement has only one point, draw a red line from the given point to the mouse position.
			if(iM[1] < 0){
				
				// Incomplete inputs are only allowed during the creation of a new angle measurement
				if(command == 3){
					plot([points[iM[0]][0], y], [points[iM[0]][1], x], "r");
				}
				
			} else {
				
				var y2,x2;
				
				// Check third point and replace it as necessary with the mouse position.
				if(iM[2] < 0){
					if(command == 3){
						y2 = y;
						x2 = x;
						color = "r";
					}
				} else {
					y2 = points[iM[2]][0];
					x2 = points[iM[2]][1];
				}
				
				// lines from the first point to the two other points
				plot([points[iM[0]][0], points[iM[1]][0]], [points[iM[0]][1], points[iM[1]][1]], "b");
				plot([points[iM[0]][0], y2], [points[iM[0]][1], x2], color);
				
				// coordinate differences
				var dy1 = points[iM[1]][0] - points[iM[0]][0];
				var dx1 = points[iM[1]][1] - points[iM[0]][1];
				var dy2 = y2 - points[iM[0]][0];
				var dx2 = x2 - points[iM[0]][1];
				
				// distances
				var s2 = Math.sqrt(dy2*dy2 + dx2*dx2);
				var s1 = Math.sqrt(dy1*dy1 + dx1*dx1);
				
				// radius of the arc (just an empirical formula)
				var r = (s1*s2)/(2*(s1+s2));

				// azimuth and angle
				var t1 = (Math.atan2(dy1, dx1) + 2*Math.PI) % (2*Math.PI);
				var t2 = (Math.atan2(dy2, dx2) + 2*Math.PI) % (2*Math.PI);
				var alpha = t2 - t1;
				
				// Draw the arc on the side, which includes a smaller angle.
				if(alpha > Math.PI){
					t2 = t2 - 2*Math.PI;
					alpha = 2*Math.PI - alpha;
				}
				if(alpha < -Math.PI){
					t1 = t1 - 2*Math.PI;
					alpha = 2*Math.PI + alpha;
				}
				alpha = Math.abs(alpha);
				
				// discrete sections on the arc
				var phi = numeric.linspace(t1, t2, Math.max(2, Math.round(20 * alpha)));

				// points on the arc
				var yA = numeric.add(points[iM[0]][0], numeric.mul(r, numeric.sin(phi)));
				var xA = numeric.add(points[iM[0]][1], numeric.mul(r, numeric.cos(phi)));

				// draw arc
				plot(yA, xA, color);
			
			}
			
		}
		
	}
	
	// visualize the distance measurements	
	for(var s = 0; s < distances.length; s++){
		
		// current measurement
		iM = distances[s];
		
		// don't draw empty measurements
		if(iM[0] > -1){
			
			// check if the measurement is complete (a total of 2 points is needed). An incomplete measurement is drawn red.
			if(iM[1] < 0){
				if(command == 2){
					plot([points[iM[0]][0], y] , [points[iM[0]][1], x] , "r");
				}
			} else {
				plot([points[iM[0]][0], points[iM[1]][0]], [points[iM[0]][1], points[iM[1]][1]],'g')
			}
			
		}
		
	}
	
	// Performe the error calculation and draw the error ellipses.
	errorCalculation();
	
	// scale factor for the error ellipses
	var L = scalebars.height * 0.60;
	var numL1 = 1.5 * numeric.sum(numeric.sqrt(numeric.getDiag(Qxx)))/Qxx.length * 1000;
	var numL = Math.round(Math.pow(Math.round(Math.sqrt(numL1 / 2)),2) / 5) * 5 * 2;
	if(numL == 0){if(numL1 > 3){numL = 4} else {numL = 2}};
	var m = L / numL * 1000;

	// draw the ellipse in the legend
	ellipseScale(numL, L);
	
	drawEllipse(m);
	
}

function plot(y,x,style,canv){
	// Function to easily draw lines and points with different symbols.
	//		Input:	- y,x: arrays with coordinates for points (y,x)
	//				- style: string with style properties
	//				- canv: canvas element
	
	if (style === undefined){
		style = "";
    }
	if (canv === undefined){
		canv = mapCtx;
    }
	
	// default
	var color = "black";
	var type = "solid";
	
	// check the style input
	var charStyle = style.split("");
	
	for(var i = 0; i < charStyle.length; i++){
		
		switch(charStyle[i]){
			case "r":
				color = "red";
				break;
			case "g":
				color = "green";
				break;
			case "b":
				color = "blue";
				break;
			case "o":
				type = "circle";
				break;
			case "^":
				type = "triangle";
				break;
		}
		
	}
	
	// create new path
	canv.beginPath();
	
	switch(type){
		
		case "solid":
			
			// connect all input points
			canv.moveTo(y[0],x[0]);
			for(var p = 1; p < y.length; p++){
				canv.lineTo(y[p],x[p]);
			}
			
			break;
			
		case "circle":
			
			// radius
			var R = 5;
			
			// draw a circle symbol for each input point
			for(var p = 0; p < y.length; p++){
				canv.moveTo(y[p]+R,x[p]);
				canv.arc(y[p],x[p],R,0,Math.PI*2,true);
			}
			
			break;
			
		case "triangle":
			
			// side length
			var s = 14;
			var dy = s/2;
			var dx = dy * Math.tan(30 * Math.PI/180);
			
			// draw a triangle symbol for each input point
			for(var p = 0; p < y.length; p++){
				canv.moveTo(y[p] + dy, x[p] + dx);
				canv.lineTo(y[p] - dy, x[p] + dx);
				canv.lineTo(y[p], x[p] - 2*dx);
				canv.lineTo(y[p] + dy, x[p] + dx);
			}
		
	}
	
	// set color and draw the path
	canv.strokeStyle = color;
	canv.stroke();	
	
}

function selectPoint(pos, pList, delta){
	// Function that finds the closest point within a certain tolerance to the input position.
	//	Input:	- pos: array with the coordinates (y,x) of the search point
	//			- pList: point list with coordinates (y,x) for each point
	//			- delta: maximal allowed tolerance between a point and the search point
	//	Output:	- pointID: ID (index of the point in the list) of the closest point
	//				(returns -1 if no point is closer than delta)
	
	// coordinates of the search point
	var y = pos[0];
	var x = pos[1];
	
	var minD = delta;
	var pointID = -1;
	
	var d, dy, dx;
	
	// check each point of the list
	for(var p = 0; p < pList.length; p++){
		
		// distance to the search point
		dy = y - pList[p][0];
		dx = x - pList[p][1];
		d = Math.sqrt(dy*dy + dx*dx)
		
		// If the current point is closer to the search point than the previous closest point, the current point becomes the new candidate.
		if(d < minD){
			minD = d;
			pointID = p;
		}
		
	}
	
	return pointID;
	
}

function clickOnPoint(mID){
	// Function to add a point to a measurement or to delete a point
	//  Input:	- mID: type of measurement or deleting
	
	switch(mID){
		
		// add distance
		case 2:
			
			// If no point has been selected yet for the distance measurement, the current point becomes the first point of the distance measurement.
			if(distances[distances.length-1][0] < 0){
                
                distances[distances.length-1][0] = pID;
                
            } else {
                
				// A point already has been selected and the current point does not correspond to the first point, the current point becomes the second point of the distance measurement
                if (pID != distances[distances.length-1][0]){
                    
					distances[distances.length-1][1] = pID;
                    
                    // Input of the distance measurement is complete, it can be added a new measurement
                    distances.push([-1, -1]);
					
					// cancel current command
					buttonClick(0);
					
                }
                
            }
			
			break;
		
		// add angle
		case 3:
		
			// Number of points already selected for the angle measurement.
            // A total of 3 points is required
            var i = angles[angles.length-1].indexOf(-1);
            
            // Check if the point exists in the current measurement already
            if(pID != angles[angles.length-1][0] && pID != angles[angles.length-1][1]){
                
                // When it does not already exist, then the point is added
                angles[angles.length-1][i] = pID;
                
				// If the measurement is complete now (3 points), create a new empty angle measurement and close the current command.
				if(i == 2){
					angles.push([-1, -1, -1]);
					buttonClick(0);
				}
                
            }

			break;
		
		// delete
		case 4:
			
			// the new point can't be deleted
			if(pID == 0){
				return;
			}
			
			// remove the point from the list
            points.splice(pID,1);

			// Also all measurements where the deleted point was involved have to be deleted.
			
            // --- Distance measurements ---
            
            var dist_temp = distances;
            distances = [];
			
			var iD;
            
            // check all distance measurements
            for(var d = 0; d < dist_temp.length; d++){
                
				// current measurement
				iD = dist_temp[d];
				
                // If the deleted point is not involved in the current measurement, the measurement can be retained
                if(iD.indexOf(pID) == -1){
                    
					// If a point in the points list is deleted, the indices of the other points change. All points that have a larger index than the deleted point will move up one position. Therefore, the indexing in the measurement has to be adjusted as necessary.

					// check both points if the inices have to be changed
					for(var i in [0, 1]){

						if(iD[i] > pID){
							iD[i]--;
						}
					}

					// add the current measurement to the new list
					distances.push(iD);
                    
                }
                
            }


            // --- Angle measurements ---
			
			// same procedure as in the distance measurements also in the angle measurements
            
            var ang_temp = angles;
            angles = [];
			
			var iA;

            for(var a = 0; a < ang_temp.length; a++){
				
				iA = ang_temp[a];

                // check if measurement can be retained
                if(iA.indexOf(pID) == -1){
                    
                    // check for all 3 points if the inices have to be changed
                    for(var i in [0,1,2]){

                        if(iA[i] > pID){
                            iA[i]--;
                        }
                    }

                    // add the current measurement to the new list
                    angles.push(iA);
                    
                }
                
            }
			
			drawGeometry();
			
			// cancel deleting function
			buttonClick(0);
					
			break;
			
	}
	
}

function movePoint(yM,xM,id){
	// Function to move a point with the mouse.
	//  Input:	- (yM,xM) coordinates of the mouse (in general they are slighty different than the coordinates of the point because of the inaccuracy of the mouse click)
	//			- id: id of the point to be moved
	
	// point coordinates
	var yP = points[id][0];
	var xP = points[id][1];
	
	// Difference between mouse click and point coordinates. This difference shall be maintained so that the moving affects more authentic.
	var dy, dx;
	dy = yP - yM;
	dx = xP - xM;
	
	// drag
	map.onmousemove = function(e){
		
		// new point coordinates
		points[id][0] = e.pageX - this.offsetLeft + dy;
		points[id][1] = e.pageY - this.offsetTop + dx;
		
		// draw the new geometry
		drawGeometry();
		// highlight the moving point in red
		if(id == 0){
			plot([points[id][0]], [points[id][1]], 'ro');
		} else {
			plot([points[id][0]], [points[id][1]], 'r^');
		}	
		
	}
	
	// drop
	map.onmouseup = function(e){
		
		map.onmousemove = mouseMove;
		map.onmouseup = null;
		
	}
		
}

function drawEllipse(m){
	// Function to draw the error ellipses of all points.
	//  Input:	- m: scale factor for the error
	
	var y,x;
	var param;
	var t,yEll,xEll;
	var theta,R,Ellipse;
	
	// all points
	for(var p = 0; p < points.length; p++){
				
		// coordinates
		y = points[p][0];
		x = points[p][1];

		// calculate the ellipse parameter
		param = calcEllipse(p);
		if(param == -1){continue};

		// Parametric representation of the ellipse
		t = numeric.linspace(0,2*Math.PI,100);
		yEll = numeric.mul(param[0], numeric.sin(t));
		xEll = numeric.mul(param[1], numeric.cos(t));

		// 2D similarity transformation
		theta = param[2] + Math.PI/2;
		R = [[Math.cos(theta), -Math.sin(theta)], [Math.sin(theta), Math.cos(theta)]];
		Ellipse = numeric.add(numeric.dot([[y], [x]], [numeric.rep([t.length],1)]) , numeric.mul(m, numeric.dot(R, [yEll, xEll])));

		// draw the ellipse
		plot(Ellipse[0],Ellipse[1],"r");
		
	}
}

function calcEllipse(id){
	// Function to calculate the parameter of an error ellipse using the variances.
	//  Input:	- id of the point whose error is to be calculated.
	//  Output:	- [A, B, theta]: ellipse parameters (or -1 if the variances of the point don't exist or are to small)
	
	// indices
	var i = 2*id;
	var j = 2*id + 1;
	
	// check if the variances exist
	if(j > Qxx.length){return -1};
		
	// variances
	var qyy = Qxx[i][i];
	var qxx = Qxx[j][j];
	var qxy = Qxx[i][j];
	
	// if the variances are close to zero the ellipse parameter are not calculated
	if(qyy <= 0.0001*0.0001 && qxx <= 0.0001*0.0001){return -1};
		
	// calculations
	var w = Math.sqrt((qyy - qxx)*(qyy - qxx) + 4*qxy*qxy);
	var A = Math.sqrt(0.5 * (qyy + qxx + w));
	var B = Math.sqrt(0.5 * (qyy + qxx - w));

	var theta = 0.5 * Math.atan2(2*qxy, (qyy - qxx));
	if(qyy == qxx){
		theta = 0;
	} else {
		theta += Math.PI/2;
	}	
	if(theta < 0){theta += Math.PI};
	if(theta > Math.PI){theta -= Math.PI};
	
	return [A, B, theta];
}

function moveSlider(slider,sId){
	// Function to handle the input of the sliders.
	//  Input:	- slider element
	//			- sId: internal id of the slider
	
	var input = slider.value/10000;
	
	// In the error calculation null values for the deviations are not allowed, so we replace them with very small deviations.
	if(input == 0){
		input = 0.00001;
	}
	
	// show the value of the chosen deviation
	switch(sId){
		case 1:
			sP = input;
			document.getElementById("lbl_sP").innerHTML = (input * 1000).toFixed(1) + " mm";
			break;
		case 2:
			sD = input;
			document.getElementById("lbl_sD").innerHTML = (input * 1000).toFixed(1) + " mm";
			break;
		case 3:
			sA = input * Math.PI/200 / mapScale;
			document.getElementById("lbl_sA").innerHTML = (input * 1000).toFixed(1) + " mgon";
			break;
	}
	
	// The errors have to be recalculated and the error ellipses have to be redrawn.
	drawGeometry();
}

function panMap(yM,xM){
	// Funktion to pan the map
	//  Input:	- (yM,XM) coordinates of the starting point of the panning
	
	map.style.cursor = "move";
	
	// drag
	map.onmousemove = function(e){
		
		// calculate displacement
		var y = e.pageX - this.offsetLeft;
		var x = e.pageY - this.offsetTop;
		var dy = y - yM;
		var dx = x - xM;
		
		// change the coordinates of all points
		for(var p = 0; p < points.length; p++){
			points[p][0] += dy;
			points[p][1] += dx;
		}
		
		// and update the coordinates of the mouse
		yM += dy;
		xM += dx;
		
		// draw changed geometry
		drawGeometry(yM,xM);
		
	}
	
	// drop
	map.onmouseup = function(e){
		
		map.onmousemove = mouseMove;
		map.onmouseup = null;
		if(command == 0){
			map.style.cursor = "default";
		}else{
			map.style.cursor = "crosshair";
		}
		
	}
	
}

function zoomMap(e){
	// Function to change the graphical representation when zooming

	// disable scrolling (http://stackoverflow.com/questions/4770025/how-to-disable-scrolling-temporarily)
	e = e || window.event;
	if (e.preventDefault)
		e.preventDefault();
	e.returnValue = false;
	
	// set a scaling factor dependig on the direction of the zoom. Prevent endless zooming
	var m;
	if(e.deltaY < 0){
		m = 1.1;
		if(scalebars.width/mapScale < 8){
			return;
		}
	} else if(e.deltaY > 0){
		m = 0.9;
		if(scalebars.width/mapScale > 20000){
			return;
		}
	} else {
		return;
	}
		
	// After the zoom the position of the mouse in the map has to be the same as before
	// Position of the mouse
	var yM = e.pageX - this.offsetLeft;
	var xM = e.pageY - this.offsetTop;

	// Transform the coordinates of all points with the scaling factor
	var Q = numeric.rep([points.length], [yM, xM, 0]); // translation
	points = numeric.add(Q, numeric.mul(m, numeric.sub(points, Q)));

	// If the geometry is scaled, also the accuracy of angle measurement must be scaled so that the relation to the distances is maintained.
	sA = sA / m;
	
	// New scaling factor of the map
	mapScale = mapScale * m;
	drawScalebar(mapScale);
	
	// Draw new geometry
	drawGeometry(yM,xM);
	
}

function drawScalebar(m){
	// Function to draw the scale bar for the map.
	//  Input:	- m: scaling factor for the map.
	
	// Some parameters of the bar:
	// heigth
	var h = 10;
	// position
	var yLeft = 20;
	var xTop = scalebars.height * 0.50;	
	// length
	var L1 = scalebars.width / 8;
	
	// calculate nominal length
	var numL1 = L1/m;
	// round the length value to a nicer number
	var z = Math.round((numL1/2).toPrecision(1));
	var numL = Math.round(numL1/z) * z;
	// scaling of the rounding
	var mNum = numL/numL1;
	// scale the bar length so the rounded value is valid
	var L = mNum * L1;
	
	// clear the old scale bar and draw the new one
	scaleCtx.clearRect(0,0,scalebars.width,scalebars.height);
	
	scaleCtx.fillStyle = "#999999";
    scaleCtx.fillRect(yLeft, xTop, L, h);
	scaleCtx.fillRect(yLeft + 2*L, xTop, L, h);

	scaleCtx.strokeStyle = "#666666";
	scaleCtx.strokeRect(yLeft, xTop, 4*L, h);
	
	scaleCtx.fillStyle = "black";
	scaleCtx.font = "12px Arial";
	scaleCtx.textAlign = "center";
	
	// labeling
	var scText;
	for(var n = 0; n < 5; n++){
		scText = (n * numL).toString();
		if(n == 4){
			scText = "        " + scText + "   [m]";
		}
		scaleCtx.fillText(scText, yLeft + n * L, xTop - 4);
	}
	
}

function ellipseScale(numA, A){
	// Function to draw a reference ellipse for comparing the error ellipses.
	//  Input:	- numA: nominal value of the semi-major axis
	//			- A: semi-major axis in pixels
	
	// Position of the ellipse
	var yCenter = scalebars.width * 0.80;
	var xCenter = scalebars.height * 0.60;

	// clear old ellipse
	scaleCtx.clearRect(yCenter - A - 1, 0, scalebars.width, scalebars.height);
	
	// Discrete points on the ellipse
	var t, yEll, xEll, Ellipse;
	t = numeric.linspace(0,2*Math.PI,100);
	yEll = numeric.mul(A, numeric.sin(t));
	xEll = numeric.mul(A/2, numeric.cos(t));

	// translation to the center point
	Ellipse = numeric.add(numeric.dot([[yCenter], [xCenter]], [numeric.rep([t.length],1)]) , [yEll, xEll]);

	// draw the ellipse
	plot(Ellipse[0],Ellipse[1],"r",scaleCtx);
	plot([yCenter, yCenter, yCenter + A], [xCenter - A/2, xCenter, xCenter], "r", scaleCtx);
	
	// labeling
	var numAstr, numBstr;
	if(numA > 9999){
		numAstr = ">9999";
		numBstr = "";
	} else {
		numAstr = numA.toString();
		numBstr = (numA/2).toString();
	}
	if(isNaN(numA)){
		numAstr = "2";
		numBstr = "1";
	}
	scaleCtx.fillStyle = "black";
	scaleCtx.font = "12px Arial";
	scaleCtx.textAlign = "left";
	scaleCtx.fillText(numAstr + "  [mm]", yCenter + A + 5, xCenter + 3);
	scaleCtx.textAlign = "center";
	scaleCtx.fillText(numBstr, yCenter, xCenter - A/2 - 4);
	
}